"""
This module is based in:

 - Database access modeule (use DAO)
 - Sanitize action (use DAO)
 - Garbage action (use DAO)
 - Pending lines proccess acction
   
   - For each pening line, execute all configured jobs.
   - Record entry into database
   - All problems are explicited in stdout.

 - This module is directly used by executables in tools directory
         
         
Tools:

 - Herramienta que de le dice los ficheros que tiene que parsear y te
   devuelve las lineas no parseadas desde la ultima ejecucion

 - Procesador de medios:
     - Se falicita un ejemplo en el de tarea de cron con fichero de bloqueo
     - Lee de un fichero de configuracion: jobs, directorios y tipo de dao a
       base de datos.
     - Por cada entrada ejecuta tareas 
     - Logea en syslog
     - jobs:

        * 
        * cp del medio a directorio temporal
        * ...

  - Tool de saneamiento: 
     - Modo forzado  o no
     - El modo forzado elimina las entradas de medios que no estan en
       el directorio
     - Revisa los medios no dados de alta
     - Revisa medios dados de alta que no existen

  - Tool de garbage

     - Limpia medios obsoletos 
     - Fichero de configuracion: base de datos, directorio?, tiempo de
       retencion


Test units

 - para los parsers
 - para los dao
 - para la fachada

"""

import pyncomings
# import pyncomings.dao
# import pyncomings.parsers
import os
import time
import re

class PyncomingsManager (object):
  import ConfigParser

  _dao={}
  _parser=None
  logfile_names=[]
  jobs=[]
  retention_days=1095

  def __init__(self,configuration_file):

    # RawConfigParser not interpolate attribute values
     cfg = self.ConfigParser.RawConfigParser()
     cfg.readfp(file(configuration_file))

     # Initialize Global options
     self.retention_days = cfg.get('global','retention_days')
 
     # Initialize database DAO
     database_type = cfg.get('database','type')
     database_file = cfg.get('database','file')
     
     # Adding DAO objects to dao dict ...
     # eval ("import %s" % database_type)
     __import__ (database_type)
     self._dao["DAOEntry"] = eval ("%s.DAOEntry('%s')" \
            % (database_type, database_file))

     # Initialize parser
     self.logfile_names.append(cfg.get('parser','log_name'))
     self.logfile_names.append(cfg.get('parser','rotate_log_name'))
     
     parser_type = cfg.get('parser','type')
     parser_mark_file_name = cfg.get('parser','mark_file_name')

     __import__ (parser_type)
     self._parser = eval ("%s.LogFileParser('%s')"
             % (parser_type, parser_mark_file_name))

     # Initialize jobs
     jobs_id = eval (cfg.get('jobs','ids')) # For example: 'job_1','
     for job_id in jobs_id:
       job_config_file=eval(cfg.get(job_id,'execute'))
       self.jobs.append (job_config_file)
    
  def sanitize_database(self,force_delete=False):
    entries = self._dao["DAOEntry"].get_entries()
    for e in entries:
      if float(e["purged_date"]) == float(0):
        try:
          file(e["path"])
        except IOError:
          # Dont file dont exist
          print "Purging %s" % e["path"]
          e["purged_date"] = time.time()
          self._dao["DAOEntry"].update_entry(e)

  def purge_garbage(self):
    entries = self._dao["DAOEntry"].get_entries()
    for e in entries:
      if float(e["purged_date"]) == float(0):
          limit_date = time.time() - (float (self.retention_days) * 86400)

          if float(e["registration_date"]) < float(limit_date):
            try:
               print "Purging %s" % e["path"]
               e["path"] = self._fix_path(e["path"], True)
               try:
                 os.remove(e["path"])
               except OSError, ex:
                 print ex

               e["purged_date"] = time.time()
               self._dao["DAOEntry"].update_entry(e)

            except Exception, ex:
               print ex


  def clean_purged(self):

    entries = self._dao["DAOEntry"].get_entries()
    for e in entries:
      # If is a entry already purged ...
      if not float(e["purged_date"]) == float(0):

          self._dao["DAOEntry"].delete_entry(e["path"])
          print "Cleaning %s purged file" % e["path"]

  def _fix_path (self,path,noescape=False):

           #XXX: Miramos que entry es un fichero valido en el sistema
           #XXX: Si no lo es, lo buscamos y devolvemos el path real en el
           #     sistema.
           # Cogemos la entrada del log -> le quitamos espacion y barras
           # bajas
           # Cogemos todos los del directorio -> les quitamos los espacios
           # y las barras
           # Los comparamos
           # Cuando machea uno, sabemos que es ESE. Cogemos una parte
           # y construimos el path completo
           if not os.path.exists(path):
               basedir_path = os.path.dirname(path)
               basename_path = os.path.basename(path)
               basename_escaped_path = basename_path.replace(" ","").replace("_","")
               try:
                 all_entries_paths = os.listdir(basedir_path)
                 for p in all_entries_paths:
                   # s = re.search("[0-9]+-[0-9]+-[0-9]+", e)
                   p_escaped = p.replace(" ","").replace("_","")
                   # id_asset = s.group()
                   if basename_escaped_path == p_escaped:
                       path = basedir_path + os.path.sep + p
                       if not noescape:
                         path = path.replace(" ", "\ ")

                       break
               except Exception:
                   pass

           return path

  def pending_lines_process(self):
    _,pending_lines = self._parser.get_pending_lines(self.logfile_names)
    # print pending_lines
    for l in pending_lines:
     # Mark line as last line processed
     self._parser.save_mark(l)

     entry = self._parser.parse_line(l)
     if entry:
       try:
         entry["retention_days"]=self.retention_days
         
         # Check if entry exist, 
         # if exist, update values
         # else, create new entry in database
         _e = self._dao["DAOEntry"].get_entry(entry["path"])
         if _e:
           self._dao["DAOEntry"].update_entry(entry)
         else:
           self._dao["DAOEntry"].create_entry(entry)
         
         for j in self.jobs:
           _ep = entry["path"]

           entry["path"] = self._fix_path(_ep)

           b = os.popen(j % entry)
           print j % entry
           for l in b.readlines():
             print l.split("\n")[0]
         
       except Exception, e:
         print e

  def create_database(self):
    self._dao["DAOEntry"].create_table()




from setuptools import find_packages

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


version = "0.3.5"

long_description = ""
try:
    long_description=file('README.md').read()
except Exception:
    pass

license = ""
try:
    license=file('LICENSE').read()
except Exception:
    pass




setup(
    name='pyncomings',
    version=version,
    author_email='saavedra.pablo@interoud.com',
    author='Pablo Saavedra Rodinho',
    description='Log entries parser',
    url='https://tracker.interoud.com/sistemas/',
    install_requires=[],
    packages=[
        'pyncomings',
        'pyncomings.dao',
        'pyncomings.parsers'
    ],
    scripts=[
        'tools/pyncomings',
        'tools/pyncomings-sanitizer',
        'tools/pyncomings-garbage-collector',
    ],
    data_files=[
        ('share/doc/pyncomings', ['doc/pyncomings.conf',
                ])
    ],


    # download_url= 'https://github.com/psaavedra/chic/zipball/master',
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    long_description=long_description,
    license=license,
    keywords = "python parser log",

)




# Authors: Pablo Saavedra
# Contact: saavedra.pablo@gmail.com

"""
This package contains ExtendedLog Parser module

Ref:
      http://doc.xarli.net/proftpd-doc/modules/mod_log.html
      http://doc.xarli.net/proftpd-doc/modules/mod_log.html#LogFormat

Example:

      ::ffff:10.121.55.149 UNKNOWN bahamas [29/Nov/2012:12:11:07 +0000] "CWD /carlos" 250 -
      ::ffff:10.121.55.149 UNKNOWN bahamas [29/Nov/2012:12:11:07 +0000] "RNTO /carlos/example.file" 250 0
"""
from pyncomings import parsers

import time

GLOBAL_CWD = ""

class LogFileParser(parsers.LogFileParser):

    def __init__(self, mark_file, jobs=[]):
        super(LogFileParser, self).__init__(mark_file, jobs)
    
    
    def parse_line(self, line):
        """
        Parse a log line.
        """
        global GLOBAL_CWD
        line_items = line.split('"')
        try:
            path = line_items[1].split()[1]
            if line_items[1].split()[0] == "RNTO":
                if path[0] != "/":
                    path = GLOBAL_CWD + "/" + path
                entry = {
                  "entry_id": None,
                  "registration_date": time.time() ,
                  "source": line_items[0].split()[0],
                  "size": line_items[2].split()[1],
                  "status": line_items[2].split()[0],
                  "command": line_items[1].split()[0],
                  "path": path,
                  "user": line_items[0].split()[2],
                  "retention_days": 0,
                  "purged_date": False
                }
                return entry
            if line_items[1].split()[0] == "CWD":
                GLOBAL_CWD = path
            else:
             return None
        except Exception:
          return None






# Authors: Pablo Saavedra
# Contact: saavedra.pablo@gmail.com

"""
This package contains DAO implementation fro SQLite3 module.
"""


from pyncomings import dao

import sqlite3




class DAOEntry(dao.DAOEntry):
    """
    entry = {
         "registration_date": None,
         "source": None,
         "size": None,
         "path": None,
         "user": None,
         "retention_days": None,
         "purged_date": None,
        }

    """

    def __init__(self, sqlite3_filename):
        self.conn = sqlite3.connect(sqlite3_filename,
                detect_types=sqlite3.PARSE_DECLTYPES|sqlite3.PARSE_COLNAMES)


    def create_table(self):
        c = self.conn.cursor()
        c.execute('''create table entries 
        (registration_date text, 
         source text, size text,
         path text primary key, user text,
         retention_days real, 
         purged_date text
        )''')
        
        self.conn.commit()
        c.close()

    def create_entry(self, entry):
        
        c = self.conn.cursor()
        c.execute('''insert into entries (registration_date, source, size,
        path, user, retention_days, purged_date) values
        (?, ?, ?, ?, ?, ?, ?)''' ,
                (entry['registration_date'], entry['source'], entry['size'],
                 entry['path'], entry['user'],
                 entry['retention_days'],entry['purged_date']
                ))
        
        self.conn.commit()
        c.close()

    def get_entry(self, path):
        c = self.conn.cursor()
        query = '''select registration_date, source, size, path, user,
        retention_days, purged_date from entries where path="%s"''' % path
        c.execute(query)
        r = c.fetchone()
        c.close()

        if r:
          res = {
           "registration_date": r[0],
           "source": r[1],
           "size": r[2],
           "path": r[3],
           "user": r[4],
           "retention_days": r[5],
           "purged_date": r[6],
          }
        else:
            res = None

        return res

    def get_entries(self, offset=None, max_items=None):
        res = []
        c = self.conn.cursor()
        if not offset:
          if not max_items:
            for r in c.execute('''select registration_date, 
            source, size, path, user, retention_days, purged_date from entries'''):
              item = {
                      "registration_date": r[0],
                      "source": r[1],
                      "size": r[2],
                      "path": r[3],
                      "user": r[4],
                      "retention_days": r[5],
                      "purged_date": r[6],
                     }
              res.append(item)
        else:
          raise NotImplementedError

        c.close()
        return res


    def update_entry(self, entry):
        c = self.conn.cursor()
        
        c.execute('''update entries set registration_date=?, source=?,
        size=?,
        user=?, retention_days=?, purged_date=?
        where path=?''',
                (entry['registration_date'], entry['source'], entry['size'],
                 entry['user'],
                 entry['retention_days'],entry['purged_date'],
                 entry['path']
                ))
        
        self.conn.commit()
        c.close()


    def delete_entry(self, path):
        c = self.conn.cursor()
        c.execute('''delete from entries where path="%s"''' % path)
        self.conn.commit()
        c.close()


    def search_entry_by_date(self, date):
        res = []
        c = self.conn.cursor()
        
        for r in c.execute('''select registration_date, 
            source, size, path, user, retention_days, purged_date from entries
            where path like "%s"''' % date):
    
              item = {
                      "registration_date": r[0],
                      "source": r[1],
                      "size": r[2],
                      "path": r[3],
                      "user": r[4],
                      "retention_days": r[5],
                      "purged_date": r[6],
                     }
              res.append(item)

        c.close()
        return res




def test():

  e1 = {
         "registration_date": "e1",
         "source": "e1",
         "size": "e1",
         "path": "e1",
         "user": "e1",
         "retention_days": 1,
         "purged_date": "e1",
        }


  e2 = {
         "registration_date": "e2",
         "source": "e2",
         "size": "e2",
         "path": "e2",
         "user": "e2",
         "retention_days": 2,
         "purged_date": "e2",
        }

  d = DAOEntry(":memory:")
  d.create_table()

  
  print "Get_entry result e1: " + str(d.get_entry(e1["path"]))

  d.update_entry(e1)
  ee = d.get_entries()
  print "Entries: " + str(ee)


  d.create_entry(e1)
  try:
    d.create_entry(e1)
    print "Error, can enter a entry with a path entry previously."
  except sqlite3.IntegrityError:
    print "OK: Integrity error success"
    

  d.create_entry(e2)

  ee = d.get_entries()
  print "Entries: " + str(ee)

  print "Entry e1: " + str(e1)
  print "Get_entry result e1: " + str(d.get_entry(e1["path"]))
  print "Entry e2: " + str(e2)
  print "Get_entry result e2: " + str(d.get_entry(e2["path"]))
  
  ee = d.search_entry_by_date(e1["registration_date"])
  print "Result for e1['registration_date'] search: " + str(ee)

  e22 = e2
  e22["size"]="SIZE22"
  d.update_entry(e22)
  
  ee = d.get_entry(e22["path"])
  print "Query result e2 (size modified): " + str(ee)
  
  d.delete_entry(e1["path"])
  ee = d.get_entry(e1["path"])
  print "Query result e1 (after deletion): " + str(ee)


if __name__ == '__main__':
    test()
